#include "olcUTIL_Geometry2D.h"
#include "TSXParser.h"
#include "olcPGEX_TransformedView.h"
#include "olcUTIL_Camera2D.h"
#include "olcPGEX_QuickGUI.h"

using namespace olc;
using namespace olc::utils;

const std::string TILESET_DIR="./Tiles/";

class TiledCollisionEditor : public olc::PixelGameEngine
{
	std::unordered_map<std::string,Tileset>tilesets;
	std::unordered_map<std::string,Renderable>images;
	std::string activeTileset;
	Quadrilateral*editingQuad=nullptr;
	Quadrilateral originalQuad;
	std::string selectedObj="";
	int editingPoint=4; //0-3 for the index we are editing within editingQuad.
	bool dragging=false;

	Renderable circle;

	TransformedView view;

public:
	TiledCollisionEditor()
	{
		sAppName = "TiledCollisionEditor";
	}

	Renderable rollingCounter;

public:
	bool OnUserCreate() override
	{
		SetFontSprite("font3.png");

		circle.Create(5,5);
		SetDrawTarget(circle.Sprite());
		Clear(BLANK);
		FillCircle({2,2},2);
		SetDrawTarget(nullptr);
		circle.Decal()->Update();

		std::string tilesetFilename=TILESET_DIR+"Basic Tileset.tsx";
		Tileset&tileset=tilesets[tilesetFilename]=TSXParser{tilesetFilename}.GetData();

		rollingCounter.Load("rollingcounter.png",nullptr,false,false);

		Renderable&tilesetImg=images[tilesetFilename];
		tilesetImg.Load(TILESET_DIR+tileset.filename);

		if(tilesets.size()==1)activeTileset=tilesetFilename;
		return true;
	}

	float hundredsDigit=0.f;
	float tensDigit=0.f;
	float onesDigit=0.f;

	float currentHealth=0;
	float targetHealth=60;

	bool OnUserUpdate(float fElapsedTime) override
	{
		Clear(VERY_DARK_BLUE);

		view.HandlePanAndZoom();
		
		const Tileset&tileset=tilesets[activeTileset];

		view.DrawDecal({0,0},images[activeTileset].Decal());

		for(int y=0;y<tileset.tilecount/tileset.columns;y++){
			for(int x=0;x<tileset.columns;x++){
				view.DrawRectDecal(vf2d{float(x),float(y)}*tileset.tilewidth,vf2d{float(tileset.tilewidth),float(tileset.tileheight)},GREY);
			}
		}

		if(editingQuad==nullptr){
			std::string selectedObj="";
		}

		for(auto&[objName,obj]:tileset.objects){
			view.DrawLineDecal(obj.bounds.pos,obj.bounds.pos+vf2d{0.f,float(obj.bounds.size.y)},YELLOW);
			view.DrawLineDecal(obj.bounds.pos,obj.bounds.pos+vf2d{float(obj.bounds.size.x),0.f},YELLOW);
			view.DrawLineDecal(obj.bounds.pos+obj.bounds.size,obj.bounds.pos+obj.bounds.size+vf2d{0.f,-float(obj.bounds.size.y)},YELLOW);
			view.DrawLineDecal(obj.bounds.pos+obj.bounds.size,obj.bounds.pos+obj.bounds.size+vf2d{-float(obj.bounds.size.x),0.f},YELLOW);

			vi2d nameTextSize=GetTextSizeProp(objName)*0.25f;
			view.GradientFillRectDecal(obj.bounds.pos,nameTextSize+vf2d{2,2},RED,{255,0,0,64},{255,0,0,64},RED);
			view.DrawStringPropDecal(obj.bounds.pos+vf2d{1.25f,1.25f},objName,BLACK,vf2d{0.25f,0.25f});
			view.DrawStringPropDecal(obj.bounds.pos+vf2d{1,1},objName,WHITE,vf2d{0.25f,0.25f});

			if(geom2d::contains(obj.bounds,view.ScreenToWorld(GetMousePos()))){
				selectedObj=objName;
			}
		}
		
		if(selectedObj.length()>0){
			for(int y=0;y<tileset.tilecount/tileset.columns;y++){
				for(int x=0;x<tileset.columns;x++){
					if(!geom2d::contains(tileset.objects.at(selectedObj).bounds,vf2d{float(x*tileset.tilewidth)+tileset.tilewidth/2,float(y*tileset.tileheight)+tileset.tileheight/2})){
						view.FillRectDecal(vf2d{float(x),float(y)}*tileset.tilewidth,vf2d{float(tileset.tilewidth),float(tileset.tileheight)},{0,0,0,128});
					}
				}
			}

			const TilesetObject&obj=tileset.objects.at(selectedObj);
			const bool EditingQuad=(editingPoint<4||dragging)&&editingQuad!=nullptr;


			auto GetSnapPoint=[&](){
				vf2d worldCoords=view.ScreenToWorld(GetMousePos());
				if(GetKey(CTRL).bHeld){
					return vf2d{round(worldCoords.x),round(worldCoords.y)};
				}
				return vf2d{round(worldCoords.x/(tileset.tilewidth/4))*tileset.tilewidth/4,round(worldCoords.y/(tileset.tileheight/4))*tileset.tileheight/4};
			};

			if(EditingQuad&&!dragging){
				(*editingQuad)[editingPoint]=GetSnapPoint();
			}else
			if(EditingQuad&&dragging){
				vf2d cursorPos=GetSnapPoint();
				vf2d initialPoint=(*editingQuad)[0];
				(*editingQuad)[1]=vf2d{cursorPos.x,initialPoint.y};
				(*editingQuad)[2]=GetSnapPoint();
				(*editingQuad)[3]=vf2d{initialPoint.x,cursorPos.y};
			}

			if(GetMouse(Mouse::LEFT).bPressed){
				#pragma region Select a point on a collision quad.
				for(const Quadrilateral&quad:obj.collisionTiles){
					for(size_t pointInd=0;const vf2d&point:quad){
						if(geom2d::line<float>(point,view.ScreenToWorld(GetMousePos())).length()<4){
							if(GetMouse(Mouse::LEFT).bPressed){
								editingPoint=pointInd;
								editingQuad=const_cast<Quadrilateral*>(&quad);
							}
							goto exitCollisionCheck;
						}
						pointInd++;
					}
				}
				exitCollisionCheck:
				#pragma endregion
				if(EditingQuad&&!dragging){
					(*editingQuad)[editingPoint]=GetSnapPoint();
					editingPoint++;
				}else{
					if(obj.collisionTiles.size()==0){
						Quadrilateral newQuad{GetSnapPoint()};
						tilesets[activeTileset].objects[selectedObj].collisionTiles.push_back(newQuad);
						dragging=true;
						editingQuad=const_cast<Quadrilateral*>(&obj.collisionTiles.back());	
						originalQuad=*editingQuad;
					}
				}
			}

			if(GetMouse(Mouse::LEFT).bReleased){
				if(EditingQuad&&dragging){
					vf2d cursorPos=GetSnapPoint();
					vf2d initialPoint=(*editingQuad)[0];
					(*editingQuad)[1]=vf2d{cursorPos.x,initialPoint.y};
					(*editingQuad)[2]=GetSnapPoint();
					(*editingQuad)[3]=vf2d{initialPoint.x,cursorPos.y};
					dragging=false;
				}else
				if(EditingQuad&&!dragging){
					(*editingQuad)[editingPoint]=GetSnapPoint();
					editingPoint=4;
				}
			}

			if(GetMouse(Mouse::RIGHT).bPressed||GetKey(ESCAPE).bPressed){
				if(EditingQuad){
					editingPoint=4;
					dragging=false;
					*editingQuad=originalQuad;
					editingQuad=nullptr;
				}
			}

			for(const Quadrilateral&quad:obj.collisionTiles){
				std::vector<vf2d>points;
				std::vector<vf2d>uvs;
				std::vector<Pixel>cols;
				points.assign(quad.begin(),quad.end());
				uvs.assign(4,{0.f,0.f});
				cols.assign(4,{255,40,40,128});
				view.DrawPolygonDecal(nullptr,points,uvs,cols);

				for(bool highlighted=false;const vf2d&point:quad){
					if(geom2d::line<float>(point,view.ScreenToWorld(GetMousePos())).length()<4&&!highlighted){
						view.DrawRotatedDecal(point,circle.Decal(),0.f,circle.Sprite()->Size()/2,{1.f,1.f},YELLOW);
						highlighted=true;
					}else{
						view.DrawRotatedDecal(point,circle.Decal(),0.f,circle.Sprite()->Size()/2,{1.f,1.f},RED);
					}
				}
			}
		}
		
		 //Font test.
		/*DrawStringDecal({0,0},"the quick brown fox jumps over the lazy dog 1234567890 !@#$%^&*()-=_+[]{}\\;':\",./<>?~`",WHITE,{1.5f,1.5f});
		DrawStringDecal({0,18},"THE QUICK BROWN FOX JUMPS OVER THE LAZY DOG 1234567890 !@#$%^&*()-=_+[]{}\\;':\",./<>?~`",WHITE,{1.5f,1.5f});
		DrawStringPropDecal({0,36},"the quick brown fox jumps over the lazy dog 1234567890 !@#$%^&*()-=_+[]{}\\;':\",./<>?~`",WHITE,{1.5f,1.5f});
		DrawStringPropDecal({0,54},"THE QUICK BROWN FOX JUMPS OVER THE LAZY DOG 1234567890 !@#$%^&*()-=_+[]{}\\;':\",./<>?~`",WHITE,{1.5f,1.5f});
		*/

		return true;
	}
};

int main()
{	
	TiledCollisionEditor demo;
	if (demo.Construct(640, 180, 4, 4))
		demo.Start();

	return 0;
}
